/*
hts.c

Copyright (C) 1999 Lars Brinkhoff.  See COPYING for terms and conditions.

hts is the server half of httptunnel.  httptunnel creates a virtual
two-way data path tunneled in HTTP requests.
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#ifdef HAVE_SYS_POLL_H
#include <sys/poll.h>
#endif
#include <sys/time.h>

#include "common.h"

typedef struct
{
  char *me;
  char *device;
  int port;
  char *forward_host;
  int forward_port;
  size_t content_length;
} Arguments;

int debug_level = 0;
FILE *debug_file = NULL;

static void
usage (FILE *f, const char *me)
{
  fprintf (f,
"Usage: %s [OPTION]... [PORT]\n"
"Listen for incoming httptunnel connections at PORT (default port is %d).\n"
"When a connection is made, I/O is redirected to the destination specified\n"
"by the --device or --forward-port switch.\n"
"\n"
"  -c, --content-length BYTES     use HTTP PUT requests of BYTES size\n"
"                                 (k, M, and G postfixes recognized)\n"
"  -d, --device DEVICE            use DEVICE for input and output\n"
#ifdef DEBUG
"  -D, --debug [LEVEL]            enable debug mode\n"
#endif
"  -f, --forward-port HOST:PORT   connect to PORT at HOST and use it for \n"
"                                 input and output\n"
"  -h, --help                     display this help and exit\n"
#ifdef DEBUG
"  -l, --logfile FILE             specify logfile for debug output\n"
#endif
"  -V, --version                  output version information and exit\n"
"\n"
"Report bugs to %s.\n",
	   me, DEFAULT_HOST_PORT, BUG_REPORT_EMAIL);
}

static void
parse_arguments (int argc, char **argv, Arguments *arg)
{
  int c;

  /* defaults */

  arg->me = argv[0];
  arg->port = DEFAULT_HOST_PORT;
  arg->device = NULL;
  arg->forward_host = NULL;
  arg->forward_port = -1;
  arg->content_length = DEFAULT_CONTENT_LENGTH;
  
  for (;;)
    {
#ifdef HAVE_GETOPT_LONG
      int option_index = 0;
      static struct option long_options[] =
      {
	{ "help", no_argument, 0, 'h' },
	{ "version", no_argument, 0, 'V' },
#ifdef DEBUG
	{ "debug", optional_argument, 0, 'D' },
	{ "logfile", required_argument, 0, 'l' },
#endif
	{ "device", required_argument, 0, 'd' },
	{ "forward-port", required_argument, 0, 'f' },
	{ "content-length", required_argument, 0, 'c' },
	{ 0, 0, 0, 0 }
      };
#endif /* HAVE_GETOPT_LONG */

#ifdef DEBUG
      static const char *short_options = "d:D:f:hl:V";
#else
      static const char *short_options = "d:f:hV";
#endif

      c = getopt_long (argc, argv, short_options,
		       long_options, &option_index);
      if (c == -1)
	break;

      switch (c)
	{
#ifdef HAVE_GETOPT_LONG
	case 0:
	  printf ("option %s", long_options[option_index].name);
	  if (optarg)
	    printf (" with arg %s", optarg);
	  printf ("\n");
	  break;
#endif /* HAVE_GETOPT_LONG */
	  
	case 'c':
	  arg->content_length = atoi_with_postfix (optarg);
	  break;

	case 'd':
	  arg->device = optarg;
	  break;

#ifdef DEBUG
	case 'D':
	  if (optarg)
	    debug_level = atoi (optarg);
	  else
	    debug_level = 1;
	  break;

	case 'l':
	  debug_file = fopen (optarg, "w");
	  if (debug_file == NULL)
	    {
	      fprintf (stderr, "%s: couldn't open file %s for writing\n",
		       arg->me, optarg);
	      exit (1);
	    }
	  break;
#endif /* DEBUG */

	case 'f':
	  name_and_port (optarg, &arg->forward_host, &arg->forward_port);
	  if (arg->forward_port == -1)
	    {
	      fprintf (stderr, "%s: you must specify a port number.\n"
		               "%s: try '%s --help' for help.\n",
		       arg->me, arg->me, arg->me);
	      exit (1);
	    }
	  break;

	case 'h':
	  usage (stdout, arg->me);
	  exit (0);

	case 'V':
	  printf ("hts (%s) %s\n", PACKAGE, VERSION);
	  exit (0);

	case '?':
	  break;

	default:
	  printf ("?? getopt returned character code 0%o ??\n", c);
	}
    }

  if (argc - 1 == optind)
    arg->port = atoi (argv[optind]);
  else if (argc - 1 > optind)
    {
      usage (stderr, arg->me);
      exit (1);
    }

  if (arg->device == NULL && arg->forward_port == -1)
    {
      fprintf (stderr, "%s: one of --device or --forward-port must be used.\n"
	               "%s: try '%s -help' for help.\n",
	       arg->me, arg->me, arg->me);
      exit (1);
    }

  if (arg->device != NULL && arg->forward_port != -1)
    {
      fprintf (stderr, "%s: --device can't be used together with "
	                   "--forward-port.\n"
	               "%s: try '%s --help' for help.\n",
	       arg->me, arg->me, arg->me);
      exit (1);
    }

  if (debug_level == 0 && debug_file != NULL)
    {
      fprintf (stderr, "%s: --logfile can't be used without --debug\n",
	       arg->me);
      exit (1);
    }

  if (((arg->device == NULL) == (arg->forward_port == -1)) ||
      arg->port == -1 ||
      ((arg->forward_host == NULL) != (arg->forward_port == -1)))
    {
      usage (stderr, arg->me);
      exit (1);
    }
}

RETSIGTYPE
sigpipe (int sig)
{
  log_debug ("caught SIGPIPE");
  signal (SIGPIPE, sigpipe);
}

int
main (int argc, char **argv)
{
  int closed;
  int fd = -1;
  Arguments arg;
  Tunnel *tunnel;

  parse_arguments (argc, argv, &arg);

  if (debug_level == 0 || debug_file != NULL)
    daemon (0, 1);

#ifdef DEBUG
  if (debug_level > 0 && debug_file == NULL)
    debug_file = stdout;
#else
  openlog ("hts", LOG_PID, LOG_DAEMON);
#endif

  log_notice ("started with arguments:");
  log_notice ("  me = %s", arg.me);
  log_notice ("  device = %s", arg.device ? arg.device : "(null)");
  log_notice ("  port = %d", arg.port);
  log_notice ("  forward_port = %d", arg.forward_port);
  log_notice ("  forward_host = %s",
	      arg.forward_host ? arg.forward_host : "(null)");
  log_notice ("  content_length = %d", arg.content_length);

  tunnel = tunnel_new_server (arg.port, arg.content_length);
  if (tunnel == NULL)
    {
      fprintf (stderr, "%s: couldn't create tunnel\n", argv[0]);
      exit (1);
    }

#ifdef DEBUG
  signal (SIGPIPE, log_sigpipe);
#else
  signal (SIGPIPE, SIG_IGN);
#endif

  for (;;)
    {
      time_t last_tunnel_write;

      log_debug ("waiting for tunnel connection");

      if (arg.device != NULL)
	{
	  fd = open_device (arg.device);
	  log_debug ("open_device (\"%s\") = %d", arg.device, fd);
	  if (fd == -1)
	    {
	      fprintf (stderr, "%s: couldn't open %s: %s\n",
		       arg.me, arg.device, strerror (errno));
	      exit (1);
	    }
	}

      if (tunnel_accept (tunnel) == -1)
	{
	  log_notice ("couldn't accept connection: %s", strerror (errno));
	  continue;
	}
      log_notice ("connected to FIXME:hostname:port");

      if (arg.forward_port != -1)
	{
	  struct sockaddr_in addr;

	  if (set_address (&addr, arg.forward_host, arg.forward_port) == -1)
	    {
	      fprintf (stderr, "%s: couldn't forward port to %s:%d: %s\n",
		       arg.me, arg.forward_host, arg.forward_port,
		       strerror (errno));
	      exit (1);
	    }

	  fd = do_connect (&addr);
	  log_debug ("do_connect (\"%s:%d\") = %d",
		 arg.forward_host, arg.forward_port, fd);
	  if (fd == -1)
	    {
	      fprintf (stderr, "%s: couldn't connect to %s:%d: %s\n",
		       arg.me, arg.forward_host, arg.forward_port,
		       strerror (errno));
	      exit (1);
	    }
	}

      closed = FALSE;
      time (&last_tunnel_write);
      while (!closed)
	{
	  struct pollfd pollfd[2];
	  int timeout;
	  time_t t;
	  int n;
      
	  pollfd[0].fd = fd;
	  pollfd[0].events = POLLIN;
	  pollfd[1].fd = tunnel_pollin_fd (tunnel);
	  pollfd[1].events = POLLIN;

	  time (&t);
	  timeout = 1000 * (30 - (t - last_tunnel_write));
	  if (timeout < 0)
	    timeout = 0;

	  log_annoying ("poll () ...");
	  n = poll (pollfd, 2, timeout);
	  log_annoying ("... = %d", n);
	  if (n == -1)
	    {
	      fprintf (stderr, "%s: poll error: %s\n",
		       arg.me, strerror (errno));
	      exit (1);
	    }
	  else if (n == 0)
	    {
	      log_verbose ("poll() timed out");
	      tunnel_padding (tunnel, 1);
	      time (&last_tunnel_write);
	      continue;
	    }

	  log_annoying ("revents[0] = %x, revents[1] = %x, POLLIN = %x",
			pollfd[0].revents, pollfd[1].revents, POLLIN);
	  handle_input ("device or port", tunnel, fd, pollfd[0].revents,
			handle_device_input, &closed);
	  handle_input ("tunnel", tunnel, fd, pollfd[1].revents,
			handle_tunnel_input, &closed);

	  if (pollfd[0].revents & POLLIN)
	    time (&last_tunnel_write);
	}

      log_debug ("closing tunnel");
      close (fd);
      tunnel_close (tunnel);
      log_notice ("disconnected from FIXME:hostname:port");
    }

  log_debug ("destroying tunnel");
  tunnel_destroy (tunnel);
 
  return 0;
}
